import * as THREE from 'https://unpkg.com/three@v0.150.0/build/three.module.js';

var renderer, scene, camera;
var chopper;

// We will use separate shaders for Gouraud and Phong shading models.
var vertexShader, fragmentShader;
var gouraudVertexShader, gouraudFragmentShader;
var phongVertexShader, phongFragmentShader;

var statusContainer; // HTML element, where we will display the shader name
var shaderMode = 1;

// We will use a point light that moves along a trajectory
var lightPosition;
var lightTrajectory;

// Viewer position in world space
var viewerPosition = new THREE.Vector3(0.0, 0.0, 8.0);

function millis() {
    return (new Date()).getTime();
}

function toRad(degree) {
    return Math.PI * 2 * degree / 360;
}

window.onLoad = async () => {
    var canvasContainer = document.getElementById('myCanvasContainer');
    var width = 800;
    var height = 500;

    // Load the Gouraud shader codes
    gouraudVertexShader = await (await fetch('vertex_gouraud.glsl')).text();
    gouraudFragmentShader = await (await fetch('fragment_gouraud.glsl')).text();

    // Load the Phong shader codes
    phongVertexShader = await (await fetch('vertex_phong.glsl')).text();
    phongFragmentShader = await (await fetch('fragment_phong.glsl')).text();

    // Status container element
    statusContainer = document.getElementById('statusContainer');

    renderer = new THREE.WebGLRenderer();
    renderer.setSize(width, height);
    canvasContainer.appendChild(renderer.domElement);

    scene = new THREE.Scene();

    // Setup the camera
    camera = new THREE.PerspectiveCamera(80, width / height, 1, 1000);
    camera.position.set(viewerPosition.x, viewerPosition.y, viewerPosition.z);
    camera.up = new THREE.Vector3(0, 1, 0);
    camera.lookAt(new THREE.Vector3(0, -5, 0));
    scene.add(camera);

    // Set the initial shader
    changeShader();

    addHangar();
    chopper = addChopper();

    // Three.js has some classes for splines. We will look more closely at curves and splines in the Basic II module.
    // For now, know that this class creates a closed curved trajectory
    lightTrajectory = new THREE.CatmullRomCurve3([
        new THREE.Vector3(0, -4, 5),
        new THREE.Vector3(-2, 0, 3),
        new THREE.Vector3(-3, 4, 6),
        new THREE.Vector3(8, 2, 9),
    ]);
    lightTrajectory.closed = true;

    draw();

    window.addEventListener('keydown', function (event) {
        if (event.code == "ArrowRight") {
            changeShader();
        }

        if (event.code == "ArrowLeft") {
            changeShader();
        }
    }, false);
}

/**
 * Function that switches the shader code.
 */
function changeShader() {
    shaderMode = (shaderMode + 1) % 2;
    if (shaderMode == 0) { // Assign the Gouraud shader
        vertexShader = gouraudVertexShader;
        fragmentShader = gouraudFragmentShader;
        statusContainer.innerHTML = 'Gouraud shader';
    } else { // Assign the Phong shader
        vertexShader = phongVertexShader;
        fragmentShader = phongFragmentShader;
        statusContainer.innerHTML = 'Phong shader';
    }

    // Traverse the scene graph and update the code for the corresponding materials.
    scene.traverse(function (object) {
        if (object.material !== undefined && object.material instanceof THREE.ShaderMaterial) {
            object.material.vertexShader = vertexShader;
            object.material.fragmentShader = fragmentShader;
            object.material.needsUpdate = true;
        }
    });
}

function draw() {
    requestAnimationFrame(draw);

    // This will sample the light trajectory curve and assign a point value at a given time to the light position.
    var m = millis() / 4000;
    lightPosition = lightTrajectory.getPoint(m - parseInt(m));

    // Traverse the scene and update the lightPosition variable value for the shaders.
    scene.traverse(function (object) {
        if (object.material !== undefined && object.material instanceof THREE.ShaderMaterial) {
            // Send the light position transformed to camera space
            object.material.uniforms.lightPosition.value = new THREE.Vector3().copy(lightPosition).applyMatrix4(camera.matrixWorldInverse);
        }
    });

    // Hopefully you have something like this for the blades rotation
    chopper.children[1].rotation.set(0, toRad((millis() / 10) % 360), 0);

    renderer.render(scene, camera);
}

function addHangar() {
    var hangar = new THREE.Object3D();
    var halfPi = Math.PI / 2;

    var leftWall = createWall(0x555555);
    leftWall.position.set(-10, 0, 0);
    leftWall.rotation.set(0, halfPi, 0);
    hangar.add(leftWall);

    var rightWall = createWall(0x333333);
    rightWall.position.set(10, 0, 0);
    rightWall.rotation.set(0, -halfPi, 0);
    hangar.add(rightWall);

    var backWall = createWall(0x444444);
    backWall.position.set(0, 0, -10);
    hangar.add(backWall);

    var ceiling = createWall(0x111111);
    ceiling.position.set(0, 10, 0);
    ceiling.rotation.set(halfPi, 0, 0);
    hangar.add(ceiling);

    var floor = createWall(0x222222);
    floor.position.set(0, -10, 0);
    floor.rotation.set(-halfPi, 0, 0);
    hangar.add(floor);

    scene.add(hangar);
}

/**
 * Hopefully you have something similar.
 * You can copy your chopper drawing code to replace this. 
 * Use a sphere as the body.
 */
function addChopper() {
    var chopper = new THREE.Object3D();
    chopper.position.set(0, -5, 0);

    var body = createSphere(0xccccee);
    body.scale.set(3, 1, 1);
    chopper.add(body);

    var blades = new THREE.Object3D();
    var blade1 = createCube(0xcceecc);
    blade1.position.set(2.7, 1.2, 0);
    blade1.scale.set(2.4, 0.1, 0.4);
    blade1.rotation.set(0, toRad(180), 0);

    blades.add(blade1);

    var blade2 = createCube(0xcceecc);
    blade2.scale.set(2.4, 0.1, 0.4);
    blade2.position.set(-2.7, 1.2, 0);
    blades.add(blade2);

    chopper.add(blades);

    scene.add(chopper);

    return chopper;
}

function createWall(colorCode) {
    var geometry = new THREE.PlaneGeometry(20, 20, 1);
    var color = new THREE.Color(colorCode);
    var material = createShaderMaterial(color);
    var wall = new THREE.Mesh(geometry, material);

    return wall;
}

function createCube(colorCode) {
    var geometry = new THREE.BoxGeometry(2, 2, 2);
    var color = new THREE.Color(colorCode);
    var material = createShaderMaterial(color);
    var cube = new THREE.Mesh(geometry, material);

    return cube;
}

function createSphere(colorCode) {
    var geometry = new THREE.SphereGeometry(1, 6, 6); // We create a sphere approximation. Radius = 1; widthSegments = heightSegments = 6.
    var color = new THREE.Color(colorCode);
    var material = createShaderMaterial(color);
    var sphere = new THREE.Mesh(geometry, material);

    return sphere;
}

function createShaderMaterial(color) {

    return new THREE.ShaderMaterial({
        uniforms: {
            // --Task--
            // There will be 2 uniforms sent to the shader:
            //  -lightPosition, v3, vec3 - the position of the light source
            //  -color, c, vec3 - the color of the current material
            
        },
        vertexShader: vertexShader,
        fragmentShader: fragmentShader
    });
}